"""
 mod_a_recs.py : Module for resolving of A and AAAA records for given
 names

 Copyright (c) 2009, Karel Slany    (slany AT fit.vutbr.cz)
                     Zdenek Vasicek (vasicek AT fit.vutbr.cz)
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:
 
     * Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
     * Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.
     * Neither the name of the organization nor the names of its
       contributors may be used to endorse or promote products derived from this
       software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.
"""

from interface import TestModule

import mod_get_dnservs
import mod_soa_recs

class Module(TestModule):
    """Gets A records for all available nameserver addresses

       The module uses data resolved by the mod_get_dnservs module. Module
       creates for each DNS server entry (each DNS server IP address - a single
       server may have multiple addresses) a resolver. Then asks by using the
       resolvers for A and/or AAAA entries - for a specific name to be
       translated.

       The obtained A and AAAA records are stored in a tree structure created
       from python dictionaries. This structure is copied into a global
       environment variable, on order to keep the results accessible for
       other modules.

       The module tests whether all the available nameservers contain the
       requested A and/or AAAA record. If a n record is found, the module
       tests whether the answer was received as an authoritative answer.
       If you are not satisfied with the outputs, you can always try to
       increase the verbosity of the module outputs.

       Be careful when creating IPv6 based resolvers. This may cause
       the application to crash on no IPv6 networks. Be also careful when
       trying to disable the IPv4 functionality.
    """

    def __init__(self, args, modenv):
        TestModule.__init__(self,args,modenv,"Get A records")
        self.dns_env = self.get_module_modenv(mod_get_dnservs)
        self.soa_env = self.get_module_modenv(mod_soa_recs)

    def execute(self):

        res = {}

        # check option
        if (not self.args.options.arecs):
            self.set_my_modenv(res)
            return True

        # create a validating resolver
        if (self.args.options.vresolver):
            vresolver = self.ldns.ldns_resolver.new_frm_file(None)
            vresolver.set_dnssec(True)
            self.verb(2, "using validating resolver")
            while vresolver.pop_nameserver():
                pass
            for ipstr in self.args.options.vresolver:
                if (not (ipstr.find(":") != -1)):
                    # ipv4
                    ip = self.ldns.ldns_rdf.new_frm_str(ipstr,
                        self.ldns.LDNS_RDF_TYPE_A)
                else:
                    # ipv6
                    ip = self.ldns.ldns_rdf.new_frm_str(ipstr,
                        self.ldns.LDNS_RDF_TYPE_AAAA)
                vresolver.push_nameserver(ip)

        self.verb(2, "A and AAAA records:")

        res["dnservs"] = {}
        domain = self.ldns.ldns_dname_new_frm_str(self.dns_env["domain"])
        resolver = self.ldns.ldns_resolver_new()

        # for each sever name
        for servname in self.dns_env["dnservs"].keys():
            res["dnservs"][servname] = {}

            # for ipv4 addresses
            if (not self.args.options.noipv4
                and ("ipv4" in self.dns_env["dnservs"][servname].keys())):
                res["dnservs"][servname]["ipv4"] = {}

                # for every address
                for addr in self.dns_env["dnservs"][servname]["ipv4"].keys():
                    res["dnservs"][servname]["ipv4"][addr] = {}

                    if (not self.soa_env["dnservs"][servname]["ipv4"][addr]["alive"]):
                        continue

                    # ask for A and AAAA records for every argument
                    self.verb(2, "A and AAAA records for", servname, "ipv4", addr)
                    for aname in self.args.options.arecs:

                        # create resolver
                        while resolver.pop_nameserver():
                            pass
                        resolver.push_nameserver(self.dns_env["dnservs"][servname]["ipv4"][addr]["rdf"])
                        if (not resolver):
                            continue

                        dreqname = self.ldns.ldns_dname_new_frm_str(aname)
                        if (not self.args.options.noipv4):
                            # ask for A
                            pkt = resolver.query(dreqname,
                                self.ldns.LDNS_RR_TYPE_A,
                                self.ldns.LDNS_RR_CLASS_IN,
                                self.ldns.LDNS_RD)
                            if (not pkt):
                                self.err(0, "Resolver response error")
                                return False
                            arec = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_A,
                                self.ldns.LDNS_SECTION_ANSWER)
                            # tries to resolve via CNAME
                            loop_counter = 0
                            redirected = False
                            q_list = []
                            q_list.append(str(dreqname))
                            name = dreqname
                            while ((not arec) and (loop_counter < 10)):
                                redirected = True
                                # answer did not come, istead test CNAME and try to redirect
                                loop_counter += 1
                                cname = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_CNAME,
                                    self.ldns.LDNS_SECTION_ANSWER)
                                if (not cname):
                                    # try to look in authority section for NS records
                                    auths = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_NS,
                                        self.ldns.LDNS_SECTION_AUTHORITY)
                                    if (auths):
                                        added_resolvers = []
                                        # resolve authorities in aditional glue
                                        for auths_rr in auths.rrs():
                                            glue = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_A,
                                                self.ldns.LDNS_SECTION_ADDITIONAL)
                                            if (glue):
                                                for glue_rr in glue.rrs():
                                                    # add a nameserver to resolver if not prevously added
                                                    if (str(glue_rr.owner()) not in added_resolvers):
                                                        self.verb(2, "ading to resolver", glue_rr.owner())
                                                        if (len(added_resolvers) == 0):
                                                            while resolver.pop_nameserver():
                                                                pass
                                                        added_resolvers.append(str(glue_rr.owner()))
                                                        resolver.push_nameserver(glue_rr.rdf(0))
                                        if (len(added_resolvers)):
                                            self.verb(2, "asking for address of", name)
                                            pkt = resolver.query(name,
                                                self.ldns.LDNS_RR_TYPE_A,
                                                self.ldns.LDNS_RR_CLASS_IN,
                                                self.ldns.LDNS_RD)
                                            arec = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_A,
                                                self.ldns.LDNS_SECTION_ANSWER)
                                    else:
                                        break
                                else:
                                    auths = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_NS,
                                        self.ldns.LDNS_SECTION_AUTHORITY)
                                    glue = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_A,
                                        self.ldns.LDNS_SECTION_ADDITIONAL)
                                    if (auths and glue):
                                        # try to move to another resolver
                                        added_resolvers = []
                                        for auths_rr in auths.rrs():
                                            for glue_rr in glue.rrs():
                                                # add a nameserver to resolver if not prevously added
                                                if (str(glue_rr.owner()) not in added_resolvers):
                                                    self.verb(2, "ading to resolver", glue_rr.owner())
                                                    if (len(added_resolvers) == 0):
                                                        while resolver.pop_nameserver():
                                                            pass
                                                    added_resolvers.append(str(glue_rr.owner()))
                                                    resolver.push_nameserver(glue_rr.rdf(0))
                                    for cname_rr in cname.rrs():
                                        name = cname_rr.rdf(0)
                                        if (str(name) in q_list):
                                            continue
                                        self.verb(2, "redirecting via cname", servname, addr, dreqname, "->", name)
                                        q_list.append(str(str(name)))
                                        pkt = resolver.query(name,
                                            self.ldns.LDNS_RR_TYPE_A,
                                            self.ldns.LDNS_RR_CLASS_IN,
                                            self.ldns.LDNS_RD)
                                        arec = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_A,
                                            self.ldns.LDNS_SECTION_ANSWER)
                                        if (arec):
                                            break
                            if (arec):
                                # parse results
                                for arec_rr in arec.rrs():
                                    if ("A" not in res["dnservs"][servname]["ipv4"][addr].keys()):
                                        res["dnservs"][servname]["ipv4"][addr]["A"] = {}
                                    if (aname not in res["dnservs"][servname]["ipv4"][addr]["A"].keys()):
                                        res["dnservs"][servname]["ipv4"][addr]["A"][aname] = []
                                    record = {}
                                    record["aa"] = pkt.aa()
                                    record["ttl"] = arec_rr.ttl()
                                    record["redirected"] = redirected
                                    record["ipv4"] = str(arec_rr.rdf(0))
                                    res["dnservs"][servname]["ipv4"][addr]["A"][aname].append(record)
                                    self.verb(2,  aname,
                                        "name:", arec_rr.owner(),
                                        "ttl:", arec_rr.ttl(),
                                        "ipv4:", arec_rr.rdf(0),
                                        "aa:", pkt.aa(), "ad:", pkt.ad())
                            
                        if (self.args.options.ipv6):
                            # ask for AAAA
                            pkt = resolver.query(dreqname,
                                self.ldns.LDNS_RR_TYPE_AAAA,
                                self.ldns.LDNS_RR_CLASS_IN,
                                self.ldns.LDNS_RD)
                            if (not pkt):
                                self.err(0, "Resolver response error")
                                return False
                            arec = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_AAAA,
                                self.ldns.LDNS_SECTION_ANSWER)
                            # tries to resolve via CNAME
                            loop_counter = 0
                            redirected = False
                            q_list = []
                            q_list.append(str(dreqname))
                            name = dreqname
                            while ((not arec) and (loop_counter < 10)):
                                redirected = True
                                # answer did not come, istead test CNAME and try to redirect
                                loop_counter += 1
                                cname = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_CNAME,
                                    self.ldns.LDNS_SECTION_ANSWER)
                                if (not cname):
                                    # try to look in authority section for NS records
                                    auths = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_NS,
                                        self.ldns.LDNS_SECTION_AUTHORITY)
                                    if (auths):
                                        added_resolvers = []
                                        # resolve authorities in aditional glue
                                        for auths_rr in auths.rrs():
                                            glue = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_A,
                                                self.ldns.LDNS_SECTION_ADDITIONAL)
                                            if (glue):
                                                for glue_rr in glue.rrs():
                                                    # add a nameserver to resolver if not prevously added
                                                    if (str(glue_rr.owner()) not in added_resolvers):
                                                        self.verb(2, "ading to resolver", glue_rr.owner())
                                                        if (len(added_resolvers) == 0):
                                                            while resolver.pop_nameserver():
                                                                pass
                                                        added_resolvers.append(str(glue_rr.owner()))
                                                        resolver.push_nameserver(glue_rr.rdf(0))
                                        if (len(added_resolvers)):
                                            self.verb(2, "asking for address of", name)
                                            pkt = resolver.query(name,
                                                self.ldns.LDNS_RR_TYPE_AAAA,
                                                self.ldns.LDNS_RR_CLASS_IN,
                                                self.ldns.LDNS_RD)
                                            arec = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_AAAA,
                                                self.ldns.LDNS_SECTION_ANSWER)
                                    else:
                                        break
                                else:
                                    auths = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_NS,
                                        self.ldns.LDNS_SECTION_AUTHORITY)
                                    glue = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_A,
                                        self.ldns.LDNS_SECTION_ADDITIONAL)
                                    if (auths and glue):
                                        # try to move to another resolver
                                        added_resolvers = []
                                        for auths_rr in auths.rrs():
                                            for glue_rr in glue.rrs():
                                                # add a nameserver to resolver if not prevously added
                                                if (str(glue_rr.owner()) not in added_resolvers):
                                                    self.verb(2, "ading to resolver", glue_rr.owner())
                                                    if (len(added_resolvers) == 0):
                                                        while resolver.pop_nameserver():
                                                            pass
                                                    added_resolvers.append(str(glue_rr.owner()))
                                                    resolver.push_nameserver(glue_rr.rdf(0))
                                    for cname_rr in cname.rrs():
                                        name = cname_rr.rdf(0)
                                        if (str(name) in q_list):
                                            continue
                                        self.verb(2, "redirecting via cname", servname, addr, dreqname, "->", name)
                                        q_list.append(str(str(cname_rr.rdf(0))))
                                        pkt = resolver.query(name,
                                            self.ldns.LDNS_RR_TYPE_AAAA,
                                            self.ldns.LDNS_RR_CLASS_IN,
                                            self.ldns.LDNS_RD)
                                        arec = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_AAAA,
                                            self.ldns.LDNS_SECTION_ANSWER)
                                        if (arec):
                                            break
                            if (arec):
                                # parse results
                                for arec_rr in arec.rrs():
                                    if ("AAAA" not in res["dnservs"][servname]["ipv4"][addr].keys()):
                                        res["dnservs"][servname]["ipv4"][addr]["AAAA"] = {}
                                    if (aname not in res["dnservs"][servname]["ipv4"][addr]["AAAA"].keys()):
                                        res["dnservs"][servname]["ipv4"][addr]["AAAA"][aname] = []
                                    record = {}
                                    record["aa"] = pkt.aa()
                                    record["ttl"] = arec_rr.ttl()
                                    record["redirected"] = redirected
                                    record["ipv6"] = str(arec_rr.rdf(0))
                                    res["dnservs"][servname]["ipv4"][addr]["AAAA"][aname].append(record)
                                    self.verb(2, aname,
                                        "name:", arec_rr.owner(),
                                        "ttl:", arec_rr.ttl(),
                                        "ipv6:", arec_rr.rdf(0),
                                        "aa:", pkt.aa(), "ad:", pkt.ad())
                        if ((("A" not in res["dnservs"][servname]["ipv4"][addr].keys()) or
                             (("A" in res["dnservs"][servname]["ipv4"][addr].keys()) and
                             (aname not in res["dnservs"][servname]["ipv4"][addr]["A"].keys()))) and
                             (("AAAA" not in res["dnservs"][servname]["ipv4"][addr].keys()) or
                             (("AAAA" in res["dnservs"][servname]["ipv4"][addr].keys()) and
                             (aname not in res["dnservs"][servname]["ipv4"][addr]["AAAA"].keys())))):
                            self.verb(2, aname, "none")
                    self.verb(2, "")

            # for ipv6 addresses
            if (self.args.options.resolver6
                and ("ipv6" in self.dns_env["dnservs"][servname].keys())):
                res["dnservs"][servname]["ipv6"] = {}

                # for every address
                for addr in self.dns_env["dnservs"][servname]["ipv6"].keys():
                    res["dnservs"][servname]["ipv6"][addr] = {}

                    if (not self.soa_env["dnservs"][servname]["ipv6"][addr]["alive"]):
                        continue

                    # ask for A and AAAA records for every argument
                    self.verb(2, "A and AAAA records for", servname, "ipv6", addr)
                    for aname in self.args.options.arecs:

                        # create resolver
                        while resolver.pop_nameserver():
                            pass
                        resolver.push_nameserver(self.dns_env["dnservs"][servname]["ipv6"][addr]["rdf"])
                        if (not resolver):
                            continue

                        dreqname = self.ldns.ldns_dname_new_frm_str(aname)
                        if (not self.args.options.noipv4):
                            # ask for A
                            pkt = resolver.query(dreqname,
                                self.ldns.LDNS_RR_TYPE_A,
                                self.ldns.LDNS_RR_CLASS_IN,
                                self.ldns.LDNS_RD)
                            if (not pkt):
                                self.err(0, "Resolver response error")
                                return False
                            arec = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_A,
                                self.ldns.LDNS_SECTION_ANSWER)
                            # tries to resolve via CNAME
                            loop_counter = 0
                            redirected = False
                            q_list = []
                            q_list.append(str(dreqname))
                            name = dreqname
                            while ((not arec) and (loop_counter < 10)):
                                redirected = True
                                # answer did not come, istead test CNAME and try to redirect
                                loop_counter += 1
                                cname = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_CNAME,
                                    self.ldns.LDNS_SECTION_ANSWER)
                                if (not cname):
                                    # try to look in authority section for NS records
                                    auths = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_NS,
                                        self.ldns.LDNS_SECTION_AUTHORITY)
                                    if (auths):
                                        added_resolvers = []
                                        # resolve authorities in aditional glue
                                        for auths_rr in auths.rrs():
                                            glue = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_AAAA,
                                                self.ldns.LDNS_SECTION_ADDITIONAL)
                                            if (glue):
                                                for glue_rr in glue.rrs():
                                                    # add a nameserver to resolver if not prevously added
                                                    if (str(glue_rr.owner()) not in added_resolvers):
                                                        self.verb(2, "ading to resolver", glue_rr.owner())
                                                        if (len(added_resolvers) == 0):
                                                            while resolver.pop_nameserver():
                                                                pass
                                                        added_resolvers.append(str(glue_rr.owner()))
                                                        resolver.push_nameserver(glue_rr.rdf(0))
                                        if (len(added_resolvers)):
                                            self.verb(2, "asking for address of", name)
                                            pkt = resolver.query(name,
                                                self.ldns.LDNS_RR_TYPE_A,
                                                self.ldns.LDNS_RR_CLASS_IN,
                                                self.ldns.LDNS_RD)
                                            arec = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_A,
                                                self.ldns.LDNS_SECTION_ANSWER)
                                    else:
                                        break
                                else:
                                    auths = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_NS,
                                        self.ldns.LDNS_SECTION_AUTHORITY)
                                    glue = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_AAAA,
                                        self.ldns.LDNS_SECTION_ADDITIONAL)
                                    if (auths and glue):
                                        # try to move to another resolver
                                        added_resolvers = []
                                        for auths_rr in auths.rrs():
                                            for glue_rr in glue.rrs():
                                                # add a nameserver to resolver if not prevously added
                                                if (str(glue_rr.owner()) not in added_resolvers):
                                                    self.verb(2, "ading to resolver", glue_rr.owner())
                                                    if (len(added_resolvers) == 0): 
                                                        while resolver.pop_nameserver():
                                                            pass
                                                    added_resolvers.append(str(glue_rr.owner()))
                                                    resolver.push_nameserver(glue_rr.rdf(0))
                                    for cname_rr in cname.rrs():
                                        name = cname_rr.rdf(0)
                                        if (str(name) in q_list):
                                            continue
                                        self.verb(2, "redirecting via cname", servname, addr, dreqname, "->", name)
                                        q_list.append(str(str(name)))
                                        pkt = resolver.query(name,
                                            self.ldns.LDNS_RR_TYPE_A,
                                            self.ldns.LDNS_RR_CLASS_IN,
                                            self.ldns.LDNS_RD)
                                        arec = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_A,
                                            self.ldns.LDNS_SECTION_ANSWER)
                                        if (arec):
                                            break
                            if (arec):
                                # parse results
                                for arec_rr in arec.rrs():
                                    if ("A" not in res["dnservs"][servname]["ipv6"][addr].keys()):
                                        res["dnservs"][servname]["ipv6"][addr]["A"] = {}
                                    if (aname not in res["dnservs"][servname]["ipv6"][addr]["A"].keys()):
                                        res["dnservs"][servname]["ipv6"][addr]["A"][aname] = []
                                    record = {}
                                    record["aa"] = pkt.aa()
                                    record["ttl"] = arec_rr.ttl()
                                    record["redirected"] = redirected
                                    record["ipv4"] = str(arec_rr.rdf(0))
                                    res["dnservs"][servname]["ipv6"][addr]["A"][aname].append(record)
                                    self.verb(2, aname,
                                        "name:", arec_rr.owner(),
                                        "ttl:", arec_rr.ttl(),
                                        "ipv4:", arec_rr.rdf(0),
                                        "aa:", pkt.aa(), "ad:", pkt.ad())
                        if (self.args.options.ipv6):
                            # ask for AAAA
                            pkt = resolver.query(dreqname,
                                self.ldns.LDNS_RR_TYPE_AAAA,
                                self.ldns.LDNS_RR_CLASS_IN,
                                self.ldns.LDNS_RD)
                            if (not pkt):
                                self.err(0, "Resolver response error")
                                return False
                            arec = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_AAAA,
                                self.ldns.LDNS_SECTION_ANSWER)
                            # tries to resolve via CNAME
                            loop_counter = 0
                            redirected = False
                            q_list = []
                            q_list.append(str(dreqname))
                            name = dreqname
                            while ((not arec) and (loop_counter < 10)):
                                redirected = True
                                # answer did not come, istead test CNAME and try to redirect
                                loop_counter += 1
                                cname = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_CNAME,
                                    self.ldns.LDNS_SECTION_ANSWER)
                                if (not cname):
                                    # try to look in authority section for NS records
                                    auths = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_NS,
                                        self.ldns.LDNS_SECTION_AUTHORITY)
                                    if (auths):
                                        added_resolvers = []
                                        # resolve authorities in aditional glue
                                        for auths_rr in auths.rrs():
                                            glue = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_AAAA,
                                                self.ldns.LDNS_SECTION_ADDITIONAL)
                                            if (glue):
                                                for glue_rr in glue.rrs():
                                                    # add a nameserver to resolver if not prevously added
                                                    if (str(glue_rr.owner()) not in added_resolvers):
                                                        self.verb(2, "ading to resolver", glue_rr.owner())
                                                        if (len(added_resolvers) == 0):
                                                            while resolver.pop_nameserver():
                                                                pass
                                                        added_resolvers.append(str(glue_rr.owner()))
                                                        resolver.push_nameserver(glue_rr.rdf(0))
                                        if (len(added_resolvers)):
                                            self.verb(2, "asking for address of", name)
                                            pkt = resolver.query(name,
                                                self.ldns.LDNS_RR_TYPE_AAAA,
                                                self.ldns.LDNS_RR_CLASS_IN,
                                                self.ldns.LDNS_RD)
                                            arec = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_AAAA,
                                                self.ldns.LDNS_SECTION_ANSWER)
                                    else:
                                        break
                                else:
                                    auths = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_NS,
                                        self.ldns.LDNS_SECTION_AUTHORITY)
                                    glue = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_AAAA,
                                        self.ldns.LDNS_SECTION_ADDITIONAL)
                                    if (auths and glue):
                                        # try to move to another resolver
                                        added_resolvers = []
                                        for auths_rr in auths.rrs():
                                            for glue_rr in glue.rrs():
                                                # add a nameserver to resolver if not prevously added
                                                if (str(glue_rr.owner()) not in added_resolvers):
                                                    self.verb(2, "ading to resolver", glue_rr.owner())
                                                    if (len(added_resolvers) == 0): 
                                                        while resolver.pop_nameserver():
                                                            pass
                                                    added_resolvers.append(str(glue_rr.owner()))
                                                    resolver.push_nameserver(glue_rr.rdf(0))
                                    for cname_rr in cname.rrs():
                                        name = cname_rr.rdf(0)
                                        if (str(name) in q_list):
                                            continue
                                        self.verb(2, "redirecting via cname", servname, addr, dreqname, "->", name)
                                        q_list.append(str(str(name)))
                                        pkt = resolver.query(name,
                                            self.ldns.LDNS_RR_TYPE_AAAA,
                                            self.ldns.LDNS_RR_CLASS_IN,
                                            self.ldns.LDNS_RD)
                                        arec = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_AAAA,
                                            self.ldns.LDNS_SECTION_ANSWER)
                                        if (arec):
                                            break
                            if (arec):
                                # parse results
                                for arec_rr in arec.rrs():
                                    if ("AAAA" not in res["dnservs"][servname]["ipv6"][addr].keys()):
                                        res["dnservs"][servname]["ipv6"][addr]["AAAA"] = {}
                                    if (aname not in res["dnservs"][servname]["ipv6"][addr]["AAAA"].keys()):
                                        res["dnservs"][servname]["ipv6"][addr]["AAAA"][aname] = []
                                    record = {}
                                    record["aa"] = pkt.aa()
                                    record["ttl"] = arec_rr.ttl()
                                    record["redirected"] = redirected
                                    record["ipv6"] = str(arec_rr.rdf(0))
                                    res["dnservs"][servname]["ipv6"][addr]["AAAA"][aname].append(record)
                                    self.verb(2, aname,
                                        "name:", arec_rr.owner(),
                                        "ttl:", arec_rr.ttl(),
                                        "ipv6:", arec_rr.rdf(0),
                                        "aa:", pkt.aa(), "ad:", pkt.ad())
                        if ((("A" not in res["dnservs"][servname]["ipv6"][addr].keys()) or
                             (("A" in res["dnservs"][servname]["ipv6"][addr].keys()) and
                             (aname not in res["dnservs"][servname]["ipv6"][addr]["A"].keys()))) and
                             (("AAAA" not in res["dnservs"][servname]["ipv6"][addr].keys()) or
                             (("AAAA" in res["dnservs"][servname]["ipv6"][addr].keys()) and
                             (aname not in res["dnservs"][servname]["ipv6"][addr]["AAAA"].keys())))):
                            self.verb(2, aname, "none")
                    self.verb(2, "")


            #

        # check for record presence on DNS machines
        answers = {}
        for aname in self.args.options.arecs:
            answers[aname] = {}
            answers[aname]["Aall"] = True
            answers[aname]["Aall_auth"] = True
            answers[aname]["Aredir"] = False
            answers[aname]["Alist"] = []
            answers[aname]["AAAAall"] = True
            answers[aname]["AAAAall_auth"] = True
            answers[aname]["AAAAredir"] = False
            answers[aname]["AAAAlist"] = []
        for servname in res["dnservs"].keys():
            if ((not self.args.options.noipv4) and
                ("ipv4" in res["dnservs"][servname].keys())):
                for ipv4 in  res["dnservs"][servname]["ipv4"].keys():

                    if (not self.args.options.noipv4):
                        if ("A" in res["dnservs"][servname]["ipv4"][ipv4].keys()):
                            for aname in self.args.options.arecs:
                                found = aname in res["dnservs"][servname]["ipv4"][ipv4]["A"].keys()
                                if (found):
                                    authoritative = True
                                    for record in res["dnservs"][servname]["ipv4"][ipv4]["A"][aname]:
                                        authoritative = authoritative and record["aa"]
                                        answers[aname]["Aredir"] = answers[aname]["Aredir"] or record["redirected"]
                                        # remember all founf ipv4
                                        if (record["ipv4"] not in answers[aname]["Alist"]):
                                            answers[aname]["Alist"].append(record["ipv4"])
                                else:
                                    authoritative = False
                                answers[aname]["Aall"] = answers[aname]["Aall"] and found
                                answers[aname]["Aall_auth"] = answers[aname]["Aall_auth"] and authoritative
                                self.verb(1, servname, ipv4, "found A", aname, ":", found)
                                self.verb(1, servname, ipv4, "found authoritative A answer for", aname, ":", authoritative)
                        else:
                            # no A records found on this server
                            self.verb(1, servname, ipv4, "no A records for", aname)
                            for aname in self.args.options.arecs:
                                answers[aname]["Aall"] = False
                                answers[aname]["Aall_auth"] = False

                    if (self.args.options.ipv6):
                        if ("AAAA" in res["dnservs"][servname]["ipv4"][ipv4].keys()):
                            for aname in self.args.options.arecs:
                                found = aname in res["dnservs"][servname]["ipv4"][ipv4]["AAAA"].keys()
                                if (found):
                                    authoritative = True
                                    for record in res["dnservs"][servname]["ipv4"][ipv4]["AAAA"][aname]:
                                        authoritative = authoritative and record["aa"]
                                        answers[aname]["AAAAredir"] = answers[aname]["Aredir"] or record["redirected"]
                                        # remember all found ipv6
                                        if (record["ipv6"] not in answers[aname]["AAAAlist"]):
                                            answers[aname]["AAAAlist"].append(record["ipv6"])
                                else:
                                    authoritative = False
                                answers[aname]["AAAAall"] = answers[aname]["AAAAall"] and found 
                                answers[aname]["AAAAall_auth"] = answers[aname]["AAAAall_auth"] and authoritative
                                self.verb(1, servname, ipv4, "found AAAA", aname, ":", found)
                                self.verb(1, servname, ipv4, "found authoritative AAAA answer for", aname, ":", authoritative)
                        else:
                            # no AAAA records found
                            self.verb(1, servname, ipv4, "no AAAA records for", aname)
                            for aname in self.args.options.arecs:
                                answers[aname]["AAAAall"] = False
                                answers[aname]["AAAAall_auth"] = False

            if (self.args.options.resolver6 and
                ("ipv6" in res["dnservs"][servname].keys())):
                for ipv6 in res["dnservs"][servname]["ipv6"].keys():

                    if (not self.args.options.noipv4):
                        if ("A" in res["dnservs"][servname]["ipv6"][ipv6].keys()):
                            for aname in self.args.options.arecs:
                                found = aname in res["dnservs"][servname]["ipv6"][ipv6]["A"].keys()
                                if (found):
                                    authoritative = True
                                    for record in res["dnservs"][servname]["ipv6"][ipv6]["A"][aname]:
                                        authoritative = authoritative and record["aa"]
                                        answers[aname]["Aredir"] = answers[aname]["Aredir"] or record["redirected"]
                                        # remember all founf ipv4
                                        if (record["ipv4"] not in answers[aname]["Alist"]): 
                                            answers[aname]["Alist"].append(record["ipv4"])
                                else:
                                    authoritative = False
                                answers[aname]["Aall"] = answers[aname]["Aall"] and found 
                                answers[aname]["Aall_auth"] = answers[aname]["Aall_auth"] and authoritative
                                self.verb(1, servname, ipv6, "found A", aname, ":", found)
                                self.verb(1, servname, ipv6, "found authoritative A answer for", aname, ":", authoritative)
                        else:
                            # no A records found on this server
                            self.verb(1, servname, ipv6, "no A records for", aname)
                            for aname in self.args.options.arecs:
                                answers[aname]["Aall"] = False
                                answers[aname]["Aall_auth"] = False

                    if (self.args.options.ipv6):
                        if ("AAAA" in res["dnservs"][servname]["ipv6"][ipv6].keys()):
                            for aname in self.args.options.arecs:
                                found = aname in res["dnservs"][servname]["ipv6"][ipv6]["AAAA"].keys()
                                if (found):
                                    authoritative = True
                                    for record in res["dnservs"][servname]["ipv6"][ipv6]["AAAA"][aname]:
                                        authoritative = authoritative and record["aa"]
                                        answers[aname]["AAAAredir"] = answers[aname]["Aredir"] or record["redirected"]
                                        # remember all found ipv6
                                        if (record["ipv6"] not in answers[aname]["AAAAlist"]):
                                            answers[aname]["AAAAlist"].append(record["ipv6"])
                                else:
                                    authoritative = False
                                answers[aname]["AAAAall"] = answers[aname]["AAAAall"] and found
                                answers[aname]["AAAAall_auth"] = answers[aname]["AAAAall_auth"] and authoritative
                                self.verb(1, servname, ipv6, "found AAAA", aname, ":", found)
                                self.verb(1, servname, ipv6, "found authoritative AAAA answer for", aname, ":", authoritative)
                        else:
                            # no AAAA records found
                            self.verb(1, servname, ipv6, "no AAAA records for", aname)
                            for aname in self.args.options.arecs:
                                answers[aname]["AAAAall"] = False
                                answers[aname]["AAAAall_auth"] = False
        # check records for attributes
        all_A = True
        all_AAAA = True
        all_A_auth = True
        all_AAAA_auth = True
        all_A_same_ip = True
        all_AAAA_same_ip = True
        some_A_redir = False
        some_AAAA_redir = False
        for aname in answers.keys():
            if (not self.args.options.noipv4):
                all_A = all_A and answers[aname]["Aall"]
                self.verb(1, "A records resolved for", aname, "on all nameservers:", answers[aname]["Aall"])
            if (self.args.options.ipv6):
                all_AAAA = all_AAAA and answers[aname]["AAAAall"]
                self.verb(1, "AAAA records resolved for", aname, "on all nameservers:", answers[aname]["AAAAall"])
            if (not self.args.options.noipv4):
                all_A_auth = all_A_auth and answers[aname]["Aall_auth"]
                if (answers[aname]["Aall"]):
                    self.verb(1, "A records for", aname, "all authoritative:", answers[aname]["Aall_auth"])
            if (self.args.options.ipv6):
                all_AAAA_auth = all_AAAA_auth and answers[aname]["AAAAall_auth"]
                if (answers[aname]["AAAAall"]):
                    self.verb(1, "AAAA records for", aname, "all authoritative:", answers[aname]["AAAAall_auth"])
            if (not self.args.options.noipv4):
                all_A_same_ip = all_A_same_ip and len(answers[aname]["Alist"]) and (len(answers[aname]["Alist"]) == 1)
                if (len(answers[aname]["Alist"]) > 1):
                    self.verb(1, "Some servers may have different or multiple A records for", aname)
            if (self.args.options.ipv6):
                all_AAAA_same_ip = all_AAAA_same_ip and len(answers[aname]["AAAAlist"]) and (len(answers[aname]["AAAAlist"]) == 1)
                if (len(answers[aname]["AAAAlist"]) > 1):
                    self.verb(1, "Some servers may have different or multiple AAAA records for", aname)
            if (not self.args.options.noipv4):
                some_A_redir = some_A_redir or answers[aname]["Aredir"]
                if (answers[aname]["Aredir"] and len(answers[aname]["Alist"])):
                    self.verb(1, "The resolving of A records for (", aname, ") has been redirected via other servers")
            if (self.args.options.ipv6):
                some_AAAA_redir = some_AAAA_redir or answers[aname]["AAAAredir"]
                if (answers[aname]["AAAAredir"] and len(answers[aname]["AAAAlist"])):
                    self.verb(1, "The resolving of AAAA records for (", aname, ") has been redirected via other servers")
            self.verb(1, "")    
        # print brief summary
        if (not self.args.options.noipv4):
            if (all_A):
                self.verb(0, "All A records successfully resolved: yes")
                self.verb(0, "")
                if (all_A_auth):
                    self.verb(0, "All A records answered by an authoritative answer: yes")
                else:
                    self.verb(0, "All A records answered by an authoritative answer: no")
                self.verb(0, "")
                if (some_A_redir):
                    self.verb(0, "The requested machine A record may be registered outside the requested zone.")
                    self.verb(0, "")
            else:
                self.verb(0, "All A records successfully resolved: no")
                self.verb(0, "")
        if (self.args.options.ipv6):
            if (all_AAAA):
                self.verb(0, "All AAAA records successfully resolved: yes")
                self.verb(0, "")
                if (all_AAAA_auth):
                    self.verb(0, "All AAAA records answered by an authoritative answer: yes")
                else:
                    self.verb(0, "All AAAA records answered by an authoritative answer: no")
                self.verb(0, "")
                if (some_AAAA_redir):
                    self.verb(0, "The requested machine AAAA record may be registered outside the requested zone.")
                    self.verb(0, "")
            else:
                self.verb(0, "All AAAA records successfully resolved: no")
                self.verb(0, "")

        self.set_my_modenv(res)
        return True

