"""
 mod_get_dnservs.py : Module, gets the list of nameservers in
 a domain

 Copyright (c) 2009, Karel Slany    (slany AT fit.vutbr.cz)
                     Zdenek Vasicek (vasicek AT fit.vutbr.cz)
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:
 
     * Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
     * Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.
     * Neither the name of the organization nor the names of its
       contributors may be used to endorse or promote products derived from this
       software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.
"""

from interface import TestModule

soa_keys = ["name-server","email-addr","sn","refresh","retry","expiry","min"]

class Module(TestModule):
    """Lists DNS machines in domain, resolves their IP addresses

       The listing is preformed by using a local resolver. The resolver is used
       to ask for the primary nameserver. Then a secondary resolver, which is
       connected directly to the primary nameserver, is being created.

       The secondary resolver is used to ask for primary nameservers via
       a NS record. Resolved names are stored in a tree structure consisting of
       python dictionaries. The data collected during the run of the module are
       moved in a global environment variable, which is accessible from all
       modules.

       When disabling the IPv4, the code in this module may crach raising an
       exception. This is caused on IPv4 networks only, where the resolver
       cannot connect to a DNS server by using only IPv6 adresses. Another
       reason for crashing may be the absence of a AAAA record for the primary
       nameserver. Or you may just forget to enable the creation of IPv6 based
       resolvers.

       At different verbosity levels the module may generate useful debugging
       output. Try to increase or decrease its verbosity if you are not satisfied
       with the printed result summary.
    """

    def __init__(self, args, modenv):
        TestModule.__init__(self,args,modenv,"DNS machines")
        # no dependencies

    def execute(self):
        # convert domain name
        #print self.args.options.dname
        res = {}
        res["domain"] = self.args.options.dname
        domain = self.ldns.ldns_dname_new_frm_str(res["domain"])

        # query domain for authoritative servers
        stat, resolver = self.ldns.ldns_resolver_new_frm_file(None)
        if (not resolver):
            raise Exception("Can't create resolver. Error:%s" % stat)
        pkt = resolver.query(domain, self.ldns.LDNS_RR_TYPE_SOA,
            self.ldns.LDNS_RR_CLASS_IN,self.ldns.LDNS_RD)
        soa = pkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_SOA,
           self.ldns.LDNS_SECTION_ANSWER)

        # end if no packer received or epty
        if ((not soa) or (not soa.rr_count())):
            self.err(0, "No SOA record, or record empty")
            return False

        soa_record = dict(zip(soa_keys,[str(a) for a in soa.rr(0).rdfs()]))
        res["primary"] = str(soa_record["name-server"])
        prim_ns = self.ldns.ldns_dname_new_frm_str(res["primary"])
        prim_addr = self.ldns.ldns_get_rr_list_addr_by_name(resolver,
            prim_ns, self.ldns.LDNS_RR_CLASS_IN, self.ldns.LDNS_RD)

        # create resolver from the primary nameserver
        subresolver = self.ldns.ldns_resolver_new()
        while subresolver.pop_nameserver():
            pass
        for prim_addr_rr in prim_addr.rrs():
            if ((not self.args.options.noipv4) and
                (prim_addr_rr.get_type() == self.ldns.LDNS_RR_TYPE_A)):
                subresolver.push_nameserver_rr(prim_addr_rr)
            if ((self.args.options.resolver6) and
                #0 and # *** remove here if you want to try it
                (prim_addr_rr.get_type() == self.ldns.LDNS_RR_TYPE_AAAA)):
                # dangerous - does not work on IPv4-only networks
                subresolver.push_nameserver_rr(prim_addr_rr)

        # ask the primary server for nameservers
        subpkt = (subresolver.query(domain, self.ldns.LDNS_RR_TYPE_NS,
            self.ldns.LDNS_RR_CLASS_IN,self.ldns.LDNS_RD))

        if (not subpkt):
            self.err(0, "Resolver response error")
            return False
        ns = subpkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_NS,
            self.ldns.LDNS_SECTION_ANSWER)

        # end if no package received or empty
        if ((not soa) or (not soa.rr_count())):
            self.err(0, "No SOA record, or record empty\n")
            return False

        self.verb(2, "Nameservers for domain:", domain)

        ns.sort()
        res["dnservs"] = {}
        for ns_rr in ns.rrs():
            # ask the primary server for ips of the nameserver
            res["dnservs"][str(ns_rr.ns_nsdname())] = {}
            
            if (not self.args.options.noipv4):
                subsubpkt = subresolver.query(ns_rr.ns_nsdname(),
                    self.ldns.LDNS_RR_TYPE_A,
                    self.ldns.LDNS_RR_CLASS_IN,self.ldns.LDNS_RD)
                subsubaddr = subsubpkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_A,
                    self.ldns.LDNS_SECTION_ANSWER)
                if (not subsubaddr):
                    # primary server does not have A record
                    res["dnservs"][str(ns_rr.ns_nsdname())]["primaryA"] = False
                    # ask resolver
                    subsubpkt = resolver.query(ns_rr.ns_nsdname(),
                        self.ldns.LDNS_RR_TYPE_A,
                        self.ldns.LDNS_RR_CLASS_IN,self.ldns.LDNS_RD)
                    subsubaddr = subsubpkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_A,
                        self.ldns.LDNS_SECTION_ANSWER)
                else:
                    # primary server has A record
                    res["dnservs"][str(ns_rr.ns_nsdname())]["primaryA"] = True
                if (subsubaddr and subsubaddr.rr_count()):
                    res["dnservs"][str(ns_rr.ns_nsdname())]["ipv4"] = {}
                    for subsubaddr_rr in subsubaddr.rrs():
                        #print subsubaddr_rr.rdf(0)
                        #print map(lambda x : int(x), str(subsubaddr_rr.rdf(0)).split('.'))
                        res["dnservs"][str(ns_rr.ns_nsdname())]["ipv4"][str(subsubaddr_rr.rdf(0))] = {}
                        self.verb(2, ns_rr.ns_nsdname(), "ipv4:", subsubaddr_rr.rdf(0))
                        # store rdf, because of a future use
                        res["dnservs"][str(ns_rr.ns_nsdname())]["ipv4"][str(subsubaddr_rr.rdf(0))]["rdf"] = subsubaddr_rr.rdf(0)
            
            if (self.args.options.ipv6):
                subsubpkt = subresolver.query(ns_rr.ns_nsdname(),
                    self.ldns.LDNS_RR_TYPE_AAAA,
                    self.ldns.LDNS_RR_CLASS_IN,self.ldns.LDNS_RD)
                subsubaddr = subsubpkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_AAAA,
                    self.ldns.LDNS_SECTION_ANSWER)
                if (not subsubaddr):
                    # primary server does not have AAAA record
                    res["dnservs"][str(ns_rr.ns_nsdname())]["primaryAAAA"] = False
                    # ask resolver
                    subsubpkt = resolver.query(ns_rr.ns_nsdname(),
                        self.ldns.LDNS_RR_TYPE_AAAA,
                        self.ldns.LDNS_RR_CLASS_IN,self.ldns.LDNS_RD)
                    subsubaddr = subsubpkt.rr_list_by_type(self.ldns.LDNS_RR_TYPE_AAAA,
                        self.ldns.LDNS_SECTION_ANSWER)
                else:
                    # primary server has AAAA record
                    res["dnservs"][str(ns_rr.ns_nsdname())]["primaryAAAA"] = True
                if (subsubaddr and subsubaddr.rr_count()):
                    res["dnservs"][str(ns_rr.ns_nsdname())]["ipv6"] = {}
                    for subsubaddr_rr in subsubaddr.rrs():
                        res["dnservs"][str(ns_rr.ns_nsdname())]["ipv6"][str(subsubaddr_rr.rdf(0))] = {}
                        self.verb(2, ns_rr.ns_nsdname(), "ipv6:", subsubaddr_rr.rdf(0))
                        # store rdf, because of a future use
                        res["dnservs"][str(ns_rr.ns_nsdname())]["ipv6"][str(subsubaddr_rr.rdf(0))]["rdf"] = subsubaddr_rr.rdf(0)

        self.verb(2, "")
       
        self.set_my_modenv(res)
        return True

